<?php

namespace Drupal\upsync;

use Drupal\Core\Config\StorageException;
use Drupal\Core\Config\StorageInterface;
use Drupal\Core\Config\UnsupportedDataTypeConfigException;
use Drupal\Core\DependencyInjection\ClassResolverInterface;
use Drupal\Core\DrupalKernelInterface;
use Drupal\Core\Entity\EntityDefinitionUpdateManagerInterface;
use Drupal\Core\Extension\Exception\ObsoleteExtensionException;
use Drupal\Core\Extension\ExtensionNameLengthException;
use Drupal\Core\Extension\MissingDependencyException;
use Drupal\Core\Extension\ModuleExtensionList;
use Drupal\Core\Extension\ModuleInstallerInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Utility\Error;

/**
 * Updates manager.
 */
class UpsyncManager implements UpsyncManagerInterface {

  use StringTranslationTrait;

  /**
   * The Drupal kernel.
   *
   * @var \Drupal\Core\DrupalKernelInterface
   */
  protected $drupalKernel;

  /**
   * The class resolver.
   *
   * @var \Drupal\Core\DependencyInjection\ClassResolverInterface
   */
  protected $classResolver;

  /**
   * The entity definition update manager.
   *
   * @var \Drupal\Core\Entity\EntityDefinitionUpdateManagerInterface
   */
  protected $entityDefinitionUpdateManager;

  /**
   * The active configuration store.
   *
   * @var \Drupal\Core\Config\StorageInterface
   */
  protected $configStorage;

  /**
   * The sync configuration store.
   *
   * @var \Drupal\Core\Config\StorageInterface
   */
  protected $configSyncStorage;

  /**
   * The module extension list.
   *
   * @var \Drupal\Core\Extension\ModuleExtensionList
   */
  protected $moduleExtensionList;

  /**
   * The module installer.
   *
   * @var \Drupal\Core\Extension\ModuleInstallerInterface
   */
  protected $moduleInstaller;

  /**
   * The logger.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $logger;

  /**
   * The result collector.
   *
   * @var \Drupal\upsync\ResultCollector
   */
  protected $resultCollector;

  /**
   * Constructs new UpdatesManager.
   *
   * @param \Drupal\Core\DrupalKernelInterface $drupal_kernel
   *   The Drupal kernel.
   * @param \Drupal\Core\DependencyInjection\ClassResolverInterface $class_resolver
   *   The class resolver.
   * @param \Drupal\Core\Entity\EntityDefinitionUpdateManagerInterface $entity_definition_update_manager
   *   The entity definition update manager.
   * @param \Drupal\Core\Config\StorageInterface $config_storage
   *   The active configuration store.
   * @param \Drupal\Core\Config\StorageInterface $config_sync_storage
   *   The sync configuration store.
   * @param \Drupal\Core\Extension\ModuleExtensionList $module_extension_list
   *   The module extension list.
   * @param \Drupal\Core\Extension\ModuleInstallerInterface $module_installer
   *   The module installer.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger factory.
   * @param \Drupal\upsync\ResultCollector $result_collector
   *   The result collector.
   */
  public function __construct(DrupalKernelInterface $drupal_kernel, ClassResolverInterface $class_resolver, EntityDefinitionUpdateManagerInterface $entity_definition_update_manager, StorageInterface $config_storage, StorageInterface $config_sync_storage, ModuleExtensionList $module_extension_list, ModuleInstallerInterface $module_installer, LoggerChannelFactoryInterface $logger_factory, ResultCollector $result_collector) {
    $this->drupalKernel = $drupal_kernel;
    $this->classResolver = $class_resolver;
    $this->entityDefinitionUpdateManager = $entity_definition_update_manager;
    $this->configStorage = $config_storage;
    $this->configSyncStorage = $config_sync_storage;
    $this->moduleExtensionList = $module_extension_list;
    $this->moduleInstaller = $module_installer;
    $this->logger = $logger_factory->get('upsync');
    $this->resultCollector = $result_collector;
  }

  /**
   * {@inheritdoc}
   */
  public function install(array $modules, $enable_dependencies = TRUE) {
    // When a new module is added to the filesystem, it may not be detected
    // immediately in certain contexts (like Drush commands, update hooks, or
    // batch processes). This is because the module list and container may still
    // be using cached data, even after flushing caches.
    // Although drupal_flush_all_caches() typically resets the module list and
    // rebuilds the container, it may not reflect these changes immediately in
    // the current request context.
    // To ensure the module is detected before attempting installation,
    // explicitly reset the module extension list and rebuild the container
    // before flushing the caches. This guarantees that module will be available
    // for installation.
    $this->moduleExtensionList->reset();
    $this->drupalKernel->rebuildContainer();
    // Flush all caches to ensure module system detects newly added modules.
    // This is critical for CLI/update hook contexts where the system may not
    // reflect new modules until caches are flushed. While expensive, this
    // ensures installation succeeds reliably.
    drupal_flush_all_caches();
    // Go for module installation now.
    try {
      $this->moduleInstaller->install($modules, $enable_dependencies);
      $this->resultCollector->addMessage($this->t('Installed modules: @list.', [
        '@list' => implode(', ', $modules),
      ]));
    }
    catch (MissingDependencyException | ObsoleteExtensionException | ExtensionNameLengthException $exception) {
      Error::logException($this->logger, $exception);
      $this->resultCollector->addError($this->t('Error installing modules. @error', [
        '@error' => $exception->getMessage(),
      ]));
    }
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function syncConfigs(array $config_items) {
    foreach ($config_items as $config_name) {
      if ($this->configSyncStorage->exists($config_name)) {
        try {
          // Read the config from file.
          $data = $this->configSyncStorage->read($config_name);
          // Write the config to the database.
          $this->configStorage->write($config_name, $data);
          $this->resultCollector->addMessage($this->t('Synced config: %config.', [
            '%config' => $config_name,
          ]));
        }
        catch (UnsupportedDataTypeConfigException | StorageException | \Exception $exception) {
          Error::logException($this->logger, $exception);
          $this->resultCollector->addError($this->t('Error syncing config %config: @error', [
            '%config' => $config_name,
            '@error' => $exception->getMessage(),
          ]));
        }
      }
    }
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function deleteConfigs(array $config_items) {
    foreach ($config_items as $config_name) {
      try {
        $this->configStorage->delete($config_name);
        $this->resultCollector->addMessage($this->t('Deleted config: %config.', [
          '%config' => $config_name,
        ]));
      }
      catch (StorageException | \Exception $exception) {
        Error::logException($this->logger, $exception);
        $this->resultCollector->addError($this->t('Error deleting config %config: @error', [
          '%config' => $config_name,
          '@error' => $exception->getMessage(),
        ]));
      }
    }
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getResults() {
    return $this->resultCollector->all();
  }

}
